#!/usr/bin/python
import numpy
import math
import superbasinscheme
import config
import eon_read
from eon_extract import _fit_box
import logging
logger = logging.getLogger('kmc')

class kmcSim:
    
    def __init__(self,statelist,initialstatelist,coarsegraining=False,dynamicsfilename="dynamics.txt",seed = None):
        self.initialstatelist = initialstatelist
        self.totaltime = 0.0
        self.timestep  =0.0
        self.numsteps = 0
        self.statelist = statelist
        self.seed = seed 
        self.procid = 0

        self.superbasining=None
        self.sbid = -1        

        #are we coarsegraining?
        if config.sb_on:
            print "Coarse graining enabled", config.sb_path     
            if config.sb_scheme == 'transition_counting':
                self.superbasining = superbasinscheme.TransitionCounting(config.sb_path, self.statelist, config.main_temperature / 11604.5, config.sb_tc_ntrans)
            elif config.sb_scheme == 'energy_level':
                self.superbasining = superbasinscheme.EnergyLevel(config.sb_path, self.statelist , config.main_temperature / 11604.5, config.sb_el_energy_increment)

        #Calculate total rate for each state
        for state in self.statelist:
            totalrate = 0.0
            cumulative_rate = [ ] 
            for proc in state['proclist']:
                totalrate += proc['rate']
                cumulative_rate.append(totalrate)

            state['totalrate'] = totalrate
            state['cumulative_rate'] = cumulative_rate[:]    
        
        #seed the prng
        print "setting seed to", self.seed
        numpy.random.seed(self.seed)
        
        #set the initial state
        if config.kmc_initial_state == -1:
            self.pickInitialState()
        else:
            self.stateid = config.kmc_initial_state
        self.previousstate = self.stateid
        
        

        #initialize dynamics file
        self.dynamicsfile=None        
        if config.kmc_write_dynamics:
            self.dynamicsfilename = config.path_dynamics
            self.dynamicsfile = open(self.dynamicsfilename,"w")
            dynamicsheader = "#%12s  %12s  %12s  %12s  %12s  %12s  %12s  %12s\n" % ('step-number', 'reactant-id', 'process-id', 'product-id', 'step-time', 'total-time', 'barrier', 'rate')
            self.dynamicsfile.write(dynamicsheader)        

#################################################################################################################################        

    def pickInitialState(self):
        
        if config.kmc_initial_random_boltzman==True:
            incrementalBoltzmannFactorList = [ ]
            factor = 0.0
            kTinv = config.main_temperature * 11604.505290
            for i in self.initialstatelist:
                factor += math.exp(- state['energy'] * kTinv )
                incrementalBoltzmannFactorList.append(factor)
            
            #draw random number in the range [0,factor) and pick an initial state
            p = numpy.random.random_sample()*factor
            istate = 0
            while ( p > incrementalBoltzmannFactorList[istate] ):
                istate+=1                           
            self.stateid = self.initialstatelist[istate]    
                
        else:    
            randomint = numpy.random.randint( len(self.initialstatelist) )        
            self.stateid = self.initialstatelist[randomint]
        logger.info("Set initial state to %d\n" % self.stateid) 

#################################################################################################################################        
        
    def randomStart(self):
        '''
        (Re)starts the kmc simulation from a randomly chosen initial state
        '''
        logger.info("Randomstart\n")
        self.pickInitialState()
        self.previousstate = self.stateid
        self.procid = 0

        self.totaltime = 0.0
        self.timestep  = 0.0
        self.numsteps  = 0
    
        #reset superbasins
        if config.sb_on and self.superbasining:
            self.superbasining.reset()
            self.sbid = -1 
            
#################################################################################################################################        

    def initializeDisplacement(self,molid):
        '''
        Calculate molecular displacement for each process 
        '''
        logger.info("initializing displacements for molecule %d\n" % molid)
        
        box = self.statelist[0]['reactant_config'].box
        #calculate molecular COM for every reactant.
        positions = [None] * len(self.statelist)
        for state in self.statelist:
            eon_read.read_molecules( state['reactant_config'], read_com = True ) 
            positions[ state['stateid'] ] = state['reactant_config'].molecules[molid]._com
            del state['reactant_config'].molecules
            
        #calculate displacement for every process and between every state 
        for statea in self.statelist:
            stateida = statea['stateid']
            statea['statedisplacement'] = { } 
            #displacement for each process:
            for proc in statea['proclist']:
                if proc['productid'] !=-1:
                    displacement = _fit_box( positions[ proc['productid'] ] - positions[ proc['reactantid'] ], box )
                    proc['displacement'] = displacement         
            #displacement from state to state:
            for stateidb in range(len(self.statelist)):
                displacement = _fit_box(positions[stateidb] - positions[ stateida ] , box) 
                statea['statedisplacement'][stateidb] = displacement 
                    
       

#################################################################################################################################        
        
    def __del__(self):
        if hasattr(self.dynamicsfile,'write'):
            self.dynamicsfile.close()


#################################################################################################################################
    
    def onestep(self):
        
        
        # The system might be in a superbasin
        if config.sb_on:
            sb = self.superbasining.get_containing_superbasin(self.stateid)
            
        #superbasin step            
        if sb:
            mean_time, reactant_stateid, product_stateid, procid, self.sbid = sb.step(self.stateid)                
        
        #normal step    
        else:
            p = numpy.random.random_sample()*self.statelist[self.stateid]['totalrate']            
            procid = 0 
            while(p > self.statelist[self.stateid]['cumulative_rate'][procid] ):
                procid+=1
            product_stateid = self.statelist[self.stateid]['proclist'][procid]['productid']
            reactant_stateid = self.stateid
            mean_time = 1.0 / self.statelist[self.stateid]['totalrate']
            self.sbid = -1

        self.timestep        = - mean_time * math.log(1.0 - numpy.random.random_sample() ) 
        self.totaltime      += self.timestep

   

        #update class
        self.numsteps +=1
        self.previousstate = self.stateid 
        self.reactant_id = reactant_stateid
        self.procid  = procid
        self.stateid = product_stateid

        #write the log    
        if config.kmc_write_dynamics:
            self.dynamicsfile.write("%12d  %12d  %12d  %12d  %12e  %12e  %12e  %12e %12d\n" % ( self.numsteps, 
                                                                                reactant_stateid, 
                                                                                procid, 
                                                                                product_stateid, 
                                                                                self.timestep, 
                                                                                self.totaltime,
                                                                                self.statelist[reactant_stateid]['proclist'][procid]['barrier'],
                                                                                self.statelist[reactant_stateid]['proclist'][procid]['rate'],
                                                                                self.sbid) )        
        

        #register the step if we are coarse graining.
        if config.sb_on and product_stateid != -1:
            self.superbasining.register_transition(reactant_stateid, product_stateid, procid)         
        

#################################################################################################################################
        
    def moleculedisplacement(self):
        '''
        returns the displament during the last step
        '''        
        displacement = numpy.zeros(3)
        displacement += self.statelist[ self.reactant_id ]['proclist'][ self.procid ]['displacement']
        if self.sbid !=-1:
            displacement += self.statelist[ self.previousstate ]['statedisplacement'][self.reactant_id]
        return displacement
        
#################################################################################################################################

   
