#!/usr/bin/python
import numpy
from collections import deque
import config
import sys
from math import pow

import logging
logger = logging.getLogger('msd')

class msdObject:
    
    def __init__(self):
        self.delta_t    = config.diffusion_delta_t
        self.max_t      = config.diffusion_max_t
        
        self.r          = numpy.zeros(3)
        self.deltar     = numpy.zeros(3)
        self.origin     = numpy.zeros(3)
        
        self.deltat           = 0.0
        self.t              = 0.0
        self.bin            = 0   
        self.num_segments   = 0     
        self.istep          = 0        
        
        self.numForcedRestarts = 0
        
        
        self.data = [] 
        t=0.0
        while t < self.max_t:
            bin = {'tleft':t,'tright':float( t+self.delta_t),'dx': 0.0,'dy': 0.0, 'msd':0.0, 'nsamp': 0}
            self.data.append(bin)
            t+=float(self.delta_t)
        
        self.trajectoryfile = None
        if config.diffusion_calc_trajectory:
            self.trajectoryfile = open( config.path_trajectory, "w")

        self.t = 0.0
        self._newSegment()            
        
    def __del__(self):
        if hasattr(self.trajectoryfile,'write'):
            self.trajectoryfile.close()
               
    def registerStep(self,displacement,timestep):
    
        self.r          += displacement
        self.deltar     += displacement
        self.t          += timestep
        self.deltat     += timestep         
        self.istep      += 1
        
        
        #check if we don't exceed max_t
        if self.deltat > self.max_t:
            self._newSegment()
            #don't register the first step (dr=0, dt = 0)
            return
            
                     
        #find bin
        new_bin = self.bin
        while( self.deltat > self.data[ new_bin ]['tright'] ):
            new_bin += 1
 
        delta_bin = new_bin - self.bin
        self.bin = new_bin

        
        #add squared displacement to the data (if there is a change in bins)
        if delta_bin > 0:
            self.data[ self.bin ]['dx']    += self.deltar[0]*self.deltar[0]
            self.data[ self.bin ]['dy']    += self.deltar[1]*self.deltar[1]
            self.data[ self.bin ]['msd']   += pow ( numpy.linalg.norm( self.deltar ) , 2 )
            self.data[ self.bin ]['nsamp'] += 1
        
        
        #write trajectory
        if config.diffusion_calc_trajectory:
            if self.istep % config.diffusion_trajectory_frequency == 0:
                self.trajectoryfile.write("%f %f %f\n" % (self.r[0],self.r[1],self.r[2]) )
        
        return      
      
      
    def _newSegment(self, exceededMaxT = True ):
        sys.stderr.write("MSD: new segment, time in previous segment = %e, exceededmaxt = %s\n" % (self.deltat, repr(exceededMaxT)) )
        logger.info("MSD: new segment, time in previous segment = %e, exceededmaxt = %s\n" % (self.deltat, repr(exceededMaxT)) )
        #if the new segment was not the result of the max_t being exceeded (i.e. a new state or substrate was entered)
        if exceededMaxT == False:
            self.numForcedRestarts +=1
    

        self.deltat          = 0.0         
        self.deltar          = numpy.zeros(3)
        self.bin             = 0
        self.num_segments    +=1
        

            
        
        
    def write(self,outfilein):
        if type(outfilein)==type(str):
            outfile = open(outfilein,"w")
        else:
            outfile = outfilein
                
                
        outfile.write("#MSD\n")
        outfile.write("#Max_t: %e\n#Delta_t: %e\n#Total simulated time: %e\n#Number of segments: %d\n" % (self.max_t, self.delta_t, self.t, self.num_segments))  
        outfile.write("#number of KMC sims: %d\n" % self.numForcedRestarts )
        outfile.write("#Average KMC time: %e\n" % ( self.t / ( max( self.numForcedRestarts,1) ) ))             
        for bin in self.data:
            if bin['nsamp']!=0:
                outfile.write("%e %e %e %e %d\n" % ( 0.5*( bin['tleft'] + bin['tright'] ), bin['msd']/bin['nsamp'], bin['dx']/bin['nsamp'], bin['dy']/bin['nsamp'], bin['nsamp'] ) )                        
        print "Wrote MSD to file\n"        
        return            
            
        
