##-----------------------------------------------------------------------------------
## eOn is free software: you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published by
## the Free Software Foundation, either version 3 of the License, or
## (at your option) any later version.
##
## A copy of the GNU General Public License is available at
## http://www.gnu.org/licenses/
##-----------------------------------------------------------------------------------

import os
import config
import numpy
import logging
logger = logging.getLogger('superbasin')


class Superbasin:
    """Class to manage super basin: calculate the mean residence time, exit probabilities, and perform Monte Carlo transitions out of the basin, """\
    """based on Novotny's Absorbing Markov Chain algorithm."""
    def __init__(self, path, id, statelist, state_list = None, get_state = None):
        #FIXME: self.states is literally a list of states, while in the superbasinscheme
        # self.states is a StateList object. Some renaming should happen.
        if state_list is None and get_state is None:
            raise ValueError('Superbasin must either have a list of states or a reference to get_state of a StateList')
        self.statelist = statelist
        self.id = int(id)
        self.path = path+str(self.id)

        #if the superbasin does not exist on disk yet, create new one. 
        if not os.path.isfile(self.path):
            self.states = state_list
            self.state_numbers = [i for i in self.states]
            self._calculate_stuff()
            self.write_data()
        #read superbasin from disk
        else:
            self.read_data()


    def pick_exit_state(self, entry_stateid):
        """Choose an exit state (state of the basin from which we will be leaving) using absorbing Markov chain theory."""
        if entry_stateid not in self.state_numbers:
            raise ValueError('Passed entry state is not in this superbasin')        
        entry_state_index = self.state_numbers.index(entry_stateid)

        probability_vector = self.probability_matrix.transpose()[entry_state_index]
        if abs(1.0-numpy.sum(probability_vector)) > 1e-3:
            logger.warning("the probability vector isn't close to 1.0")
            logger.warning('probability_vector ' + str(probability_vector) + " " + str(numpy.sum(probability_vector)))
        probability_vector /= numpy.sum(probability_vector)
        
        u = numpy.random.random_sample()
        p = 0.0
        for i in range(len(self.states)):
            p += probability_vector[i]
            if p>u:
                exit_state_index = i 
                break
        else:
            logger.warning("Warning: failed to select exit state. p = " + str(p))
        time = self.mean_residence_times[entry_state_index]
        return time, exit_state_index

    def test(self, entry_state):
        n=len(self.mean_residence_times)
        p=[0]*n
        m=10
        for i in range(m):
            time, exit_state_index=self.pick_exit_state(entry_state)
            p[exit_state_index]+=1
        for i in range(n):
            p[i]=float(p[i])/m
        print 'pick up p:', p
        return time, exit_state_index


    def step(self, entry_stateid):
        """Perform a Monte Carlo transition: leave the basin."""\
        """The function returns a residence time as well as information to indenfity what saddle point was to leave the basin,"""\
        """from what state and to what state the system is moving to."""
        time, exit_state_index = self.pick_exit_state(entry_stateid)
        assert(time >= 0.0)
        exit_stateid = self.state_numbers[exit_state_index]
        exit_state = self.statelist[exit_stateid]

        # Make a rate table for all the exit state.  All processes are 
        # needed as the might be a discrepancy in time scale
        # and it might be dangerous to weed out low rate events
        rate_table = []
        ratesum = 0.0
        process_table = exit_state['proclist']

        # Determine all process OUT of the superbasin (leaving from exit_state)
        for proc in process_table:
            if proc['productid'] not in self.state_numbers:
                rate_table.append([proc['procid'], proc['rate']])
                ratesum += proc['rate']
        
        # picks the process to leave the superbasin
        p = 0.0
        u = numpy.random.random_sample()
        for i in range(len(rate_table)):
            p += rate_table[i][1]/ratesum
            if p>=u:
                exit_proc_id = rate_table[i][0]
                break
        else:
            logger.warning("Warning: failed to select rate. p = " + str(p))
        
        # When requesting the product state the process 
        # gets added to the tables of events for both the forward 
        # and reverse process
        product_stateid = process_table[exit_proc_id]['productid']

        return time, exit_stateid, product_stateid, exit_proc_id, self.id

    def contains_state(self, stateid):
        return stateid in self.state_numbers

    def _calculate_stuff(self):
        """Build the transient and recurrent matrices."""\
        """Calculate the fundamental matrix in order to be able to calculate the mean resisdence time"""\
        """and exit probablities any initial distribution."""
        #The i'th component of the recurrent vector contains the sum of all rates leaving state i (which 
        #is inside the composite) and entering a state which is not in the superbasin
        recurrent_vector = numpy.zeros(len(self.states))

        #The i'th diagonal component of the transient matrix contains minus the sum of _all_ rates 
        #from processes leaving state i to any other state (whether inside the composite or not).
        #the offdiagonal [i][j] components of the transient matrix contains the rate from state 
        #i (inside the superbasin) to state j (also in the superbasin)
        transient_matrix= numpy.zeros((len(self.states), len(self.states)))
        
        for i in range(len(self.states)):
            proc_table = self.statelist[ self.state_numbers[i] ]['proclist']
            #proc_table = self.states[i].get_process_table()
            for process in proc_table:
                
                #process is leaving the superbasin 
                if process['productid']==-1 or process['productid'] not in self.state_numbers: 
                    recurrent_vector[i] += process['rate']
                    
                #process remains in superbasin    
                else:
                    j = self.state_numbers.index(process['productid'])
                    #columns and rows interchanged as compared to theory?
                    transient_matrix[j][i] += process['rate']
                    
                transient_matrix[i][i] -=process['rate']
        
        #Calculate mean residence time
        
        #Fundamental matrix is the inverse of the transient matrix T (not the inverse of (I-T) )
        fundamental_matrix = numpy.linalg.inv(transient_matrix)
     
        #mean_residence_times contains the lifetime of state i in the composite state.
        self.mean_residence_times = numpy.zeros(len(self.states))
        #the probability matrix contains on the [i][j]'th position the probability of leaving 
        #the superbasin from state j, given that the system entered the superbasin from state i (or vice versa ;) )
        self.probability_matrix = numpy.zeros((len(self.states), len(self.states)))
        
        for i in range(len(self.states)):
            for j in range(len(self.states)):
                self.mean_residence_times[j] -= fundamental_matrix[i][j]
                self.probability_matrix[i][j] = -recurrent_vector[i]*fundamental_matrix[i][j]

        #print "recurrent vector:"
        #print recurrent_vector
        #print ""
        #print "Probability matrix:"
        #print self.probability_matrix
        #print "\n"

        for i in self.probability_matrix.transpose():
            if abs(1.0-i.sum()) > 1e-3:
                logger.warning('Probability matrix has row which does not add up to 1')
                logger.warning('Row: %s' % str(i))
                logger.warning('Transient matrix:\n%s' % str(transient_matrix))
                logger.warning('Recurrent vector:\n%s' % str(recurrent_vector))
                logger.warning('Fundamental matrix:\n%s' % str(fundamental_matrix))
            


    def write_data(self):
        logger.debug('saving data to %s' %self.path)
        f = open(self.path, 'w')
        for i in [self.state_numbers, self.mean_residence_times, self.probability_matrix.ravel()]:
            for j in i:
                print >> f, repr(j), 
            print >> f
        f.close()


    def read_data(self):
        logger.info('reading data from %s' % self.path)
        f = open(self.path, 'r')
        
        
        self.state_numbers = []
        for i in f.readline().rstrip().split():
            self.state_numbers.append(int(i)) 
        self.states = self.state_numbers[:]
        self.mean_residence_times = []
        for i in f.readline().rstrip().split():
            self.mean_residence_times.append(numpy.float64(i)) 
        pmat = [] 
        for i in f.readline().rstrip().split():
            pmat.append(numpy.float64(i))
        self.probability_matrix = numpy.array(pmat).reshape((len(self.states), len(self.states)))
        f.close()        

    def delete(self, storage=None):
        if storage is None:
            logger.info('deleting %s' % self.path)
            os.remove(self.path)
        else:
            logger.info('storing %s' % self.path)
            path_storage = storage+str(self.id)
            os.rename(self.path, path_storage)
        
        self.states = None
        self.state_numbers=None
        self.probability_matrix = None
        self.mean_residence_times = None
    
