##-----------------------------------------------------------------------------------
## eOn is free software: you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published by
## the Free Software Foundation, either version 3 of the License, or
## (at your option) any later version.
##
## A copy of the GNU General Public License is available at
## http://www.gnu.org/licenses/
##-----------------------------------------------------------------------------------

import os
import logging
import config
logger = logging.getLogger('superbasinscheme')

import superbasin
import config
import shutil

class SuperbasinScheme:
    ''' This poorly-named class handles keeping track of which states belong
    to which superbasins, the SuperBasin object of those superbasins, and 
    superbasining criteria. It also expands and merges superbasins '''

    def __init__(self, superbasin_path, states, kT):
        self.path = superbasin_path
        self.path_storage = os.path.join(superbasin_path,"storage/")
        self.path_old = os.path.join(superbasin_path,"old/")
        self.states = states
        self.kT = kT
        
        if not os.path.isdir(self.path):
            logger.warning('superbasin path does not exist, creating %s' % self.path)
            os.makedirs(self.path)
            if config.debug_store_superbasins:
                os.makedirs(self.path_storage)
                os.makedirs(self.path_old)
            
        else:    
            dirs = os.listdir(self.path)
            if len(dirs)!=0:
                res = raw_input("Superbasin dir is not empty, delete contents? (i.e. start with new superbasins) (y/N)").lower()  
                if res[0]=='y':
                    self.clearDirs()   
                    if config.debug_store_superbasins:                                 
                        os.makedirs(self.path_storage)
                        os.makedirs(self.path_old)                         
                else: 
                    exit(0)                    
            
            
        self.superbasins = []
        self.resetCount = 0
        self.next_sb_num = 0
        dirs = os.listdir(self.path)        
        for i in dirs:
            if i != 'storage' and i != 'old' and i[:6]!='level_':
                self.next_sb_num = max(self.next_sb_num, int(i))
                self.superbasins.append(superbasin.Superbasin(self.path, i, self.states, get_state = True) )
        self.next_sb_num += 1
        self.read_data()        
        self.resetCount = 0

    def get_containing_superbasin(self, stateid):
        for sb in self.superbasins:
            if stateid in sb.state_numbers:
                return sb
        return None              
        #for i in self.superbasins:
        #    if i.contains_state(stateid):
        #        return i
        #return None

    def make_basin(self, merge_states):

        #first determine how many states will be in the new superbasin
        if config.sb_max_size:
            numstates = 0
            for stateid in merge_states:
                sb = self.get_containing_superbasin(stateid)
                if sb is None:
                    numstates += 1
                else:
                    numstates += len(sb.states)

            if numstates>config.sb_max_size:
                return
        

        new_sb_states = []
        for stateid in merge_states:
            sb = self.get_containing_superbasin(stateid)
            if sb is None:
                if stateid not in new_sb_states:
                    new_sb_states.append(stateid)
            else:
                for j in sb.states:
                    if j not in new_sb_states:
                        new_sb_states.append(j)
                # keep basins to analyze data
                if config.debug_store_superbasins:
                    sb.delete(self.path_storage)
                else:
                    sb.delete()

                self.superbasins.remove(sb)
                del(sb)
                
        #self.states.connect_states(new_sb_states) #XXX:This should ensure detailed balance
        #However, it will likely be very slow. We should be able to do without it.
        #Also, if confidence is changed and new processes are found, the superbasin
        #will ignore these new processes.
        logger.info("Creating superbasin with states " + str([i for i in new_sb_states]))
        self.superbasins.append(superbasin.Superbasin(self.path, self.next_sb_num, self.states, state_list = new_sb_states)) 


        self.next_sb_num += 1
    
    def register_transition(self, start_state, end_state, proc_id_out):
        raise NotImplementedError()

    def write_data(self):
        raise NotImplementedError()
    
    def read_data(self):
        raise NotImplementedError()

    def reset(self):
        raise NotImplementedError()

    def __del__(self):
        self.write_data()


class TransitionCounting(SuperbasinScheme):
    ''' Implements the transition counting scheme for superbasining '''
    
    def __init__(self, superbasin_path, states, kT, num_transitions):
        self.num_transitions = num_transitions
        SuperbasinScheme.__init__(self,superbasin_path, states, kT)

    def register_transition(self, start_state, end_state, proc_id_out):
        logger.debug('Registering transitions')
        
        if start_state == end_state:
            return
        
        start_count = self.get_count(start_state)
        if end_state not in start_count:
            start_count[end_state] = 0
        start_count[end_state] += 1
        
        if start_count[end_state] >= self.num_transitions:
            logger.debug( "Making basin....")
            self.make_basin([start_state, end_state])

    

    def write_data(self):
        logger.debug('writing')
        for start_state in self.count:
            data_path = os.path.join(start_state.path, config.sb_state_file)
            f = open(data_path, 'w')
            for end_state in self.count[start_state]:
                print >> f, end_state.number, self.count[start_state][end_state]
            f.close()

    def read_data(self):
        self.count = {}

    def get_count(self, state):
        try:
            return self.count[state]
        except:
            data_path = os.path.join(state.path, config.sb_state_file)
            self.count[state] = {}
            if os.path.isfile(data_path):
                f = open(data_path, 'r')
                for i in f:
                    i = i.strip().split()
                    self.count[state][self.states.get_state(int(i[0]))] = int(i[1]) 
                f.close()
            return self.count[state]
            
        
   

class EnergyLevel(SuperbasinScheme):

    def __init__(self, superbasin_path, states, kT, energy_increment):
        self.energy_increment = energy_increment
        self.levels = {}
        SuperbasinScheme.__init__(self,superbasin_path, states, kT)

    def get_energy_increment(self, e_min, e_global_min, e_saddle):
        assert(e_min >= e_global_min)
        #return self.energy_increment
        return self.energy_increment * (e_saddle - e_min) / (e_saddle - e_global_min)

    def get_statelist(self, stateid):
        sb = self.get_containing_superbasin(stateid)
        if sb:
            return sb.states
        else:
            return [ stateid ]           

    #start_state and end_state are the non-sb ids. 
    def register_transition(self, start_stateid, end_stateid, proc_id_out):
        '''Increments the energy level of the end state or sets it equal to the energy 
           of the end_state if it hasn't been visited before.'''
        #error
        if start_stateid == end_stateid:
            return

        #if the start state does not have an energy level yet, we set it to the energy of the state. 
        if start_stateid not in self.levels.keys():
            self.levels[start_stateid] = self.states[start_stateid]['energy']
        
        #if the end state does not have an energy level yet, set it to the energy of the state. 
        if end_stateid not in self.levels.keys():
            self.levels[end_stateid] = self.states[end_stateid]['energy']

	    #get statelists of the start and end states 
        #(may also just be [start] or [end] if they are not yet contained in a superbasin)
        start_statelist = self.get_statelist(start_stateid)
        end_statelist   = self.get_statelist(end_stateid)

        e_min        = min ( self.states[stateid]['energy'] for stateid in end_statelist )
        e_global_min = min ( [ e_min ] + [ self.states[stateid]['energy'] for stateid in start_statelist ] )


        #determine the barrier
        try:
            proc = self.states[start_stateid]['proclist'][proc_id_out]
        except: 
            logger.warning("Start and end state have no direct connection")
            return                
        
        barrier = proc['barrier']
        saddle_energy = proc['saddle_energy']


        #increment the level of the end_state. 
        increment = self.get_energy_increment(e_min, e_global_min, saddle_energy)
        for i in end_statelist: 
            self.levels[i] += increment


        #merge states if one of the two levels is higher than the saddle energy
        largest_level = max(self.levels[start_stateid], self.levels[end_stateid])

        if largest_level > saddle_energy:
            self.levels[start_stateid] = largest_level 
            self.levels[end_stateid] = largest_level 
            for i in end_statelist: #this may also be just [ end_state ]
                self.levels[i] = largest_level
            self.make_basin([start_stateid, end_stateid])


    def clearDirs(self):
        '''
        Delete everything in self.path
        '''
        print "clearing dir"
        dirs = os.listdir(self.path)
        for i in dirs:
            if i=='storage' or i=='old':
                shutil.rmtree( os.path.join(self.path,i) )
            else:
                os.remove( os.path.join(self.path,i) )
        

    def reset(self):
        logger.debug('resetting')
        self.write_data()
        for sb in self.superbasins:
            sb.write_data()

        if config.debug_store_superbasins:            
            olddir = os.path.join(self.path_old,str(self.resetCount))
            os.makedirs(olddir)
            dirs = os.listdir(self.path) 
            #copy for storage
            for i in dirs:
                if i!="old":
                    fromdir = os.path.join(self.path,i)
                    todir   = os.path.join(olddir,i)
                    shutil.copy(fromdir,todir)            

        #remove all from current dir.
        dirs = os.listdir(self.path)        
        for i in dirs:
            if i!="old":
                path = os.path.join(self.path,i)
                os.remove(path)

        if config.debug_store_superbasins:                
            os.makedirs(self.path_storage)        
        self.superbasins = [] 
        self.resetCount +=1
        #self.next_sb_num = 0
        self.levels = {}

    def read_data(self):
        logger.debug('reading data')
        for state in self.states:
            data_path = os.path.join(self.path,"level_"+str(state['stateid'] ) )
            if os.path.isfile(data_path):
                f = open(data_path, 'r')
                self.levels[ state['stateid'] ] = float(f.read().strip())
                f.close()

    def write_data(self):
        logger.debug('writing')
        for i in self.levels.keys():
            data_path = os.path.join(self.path,"level_"+str(i))
            f = open(data_path, 'w')
            print >> f, "%f\n" % self.levels[i]
            f.close()

